// ignore_for_file: public_member_api_docs

import 'dart:developer';

import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_support_for_wearcardiot/'
    'healthrian_ble_support_for_wearcardiot.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

mixin Ecg {
  /// [maxBufferStreamEcg] decides a maximum length of buffered Stream.
  ///
  /// For 40 seconds buffer and 250 sampling rate,
  /// It uses 40 * 250.
  int maxBufferStreamEcg = 40 * 250;

  /// [lengthBufferStreamEcg] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  ///
  /// For 30 seconds buffer and 250 sampling rate,
  /// It uses 30 * 250.
  int lengthBufferStreamEcg = 30 * 250;

  /// [screenBufferStreamingEcg] decides a length of buffered Stream,
  /// which is shown to screen.
  ///
  /// For 5 seconds buffer and 250 sampling rate,
  /// It uses 5 * 250 by default.
  int screenBufferStreamingEcg = 5 * 250;

  /// [refreshBufferEcg] decides a length to be hidden after refreshed.
  ///
  /// For 0.5 seconds range to be hidden and 250 sampling rate,
  /// It use 250 ~/ 2 by default.
  int refreshBufferEcg = 250 ~/ 2;

  /// [isFilteredEcg] decides whether ecg signal is filtered.
  bool isFilteredEcg = true;

  /// [isCompressedEcg] decides whether ecg signal is compressed.
  ///
  /// If it is `true`, ecg signal is decompressed by LZMA algorithm.
  bool isCompressedEcg = false;

  final _dataFilter = DataFilter(2);

  final _adaptiveHR = AdaptiveHR(0);

  bool _isPaused = false;

  bool _isMeasuring = false;

  /// [skipSampleEcg] is the number of samples to be skipped.
  int skipSampleEcg = 1;

  /// [skipPacketEcg] is the number of packet to be skipped.
  int skipPacketEcg = 1;

  // final _channelCountEcg = 1;
  final _sampleCountEcg = 60;

  // int get _totalCountOxi => _channelCountEcg * _sampleCountEcg;
}

extension BLEEcg on BLEManager {
  /// [startMeasurementEcg] sends read_ecg_packet to device,
  ///  and then, it registers all ecg streams of [streams].
  Future<void> startMeasurementEcg() async {
    await _sendCommandReadEcg();

    // final isRunnable = await _checkValidationEcg();
    // if (!isRunnable) return;

    streams
      ..rawStreamEcg = _getStreamEcg().broadcast
      ..decompressedStreamEcg = _getDecompressedStreamEcg().broadcast
      ..recentStreamEcg = _getRecentStreamEcg().broadcast
      ..leadStateStreamEcg = _getLeadStateStream().broadcast
      ..heartRateStreamEcg = _getHeartRateStream().broadcast
      ..bufferedStreamEcg = _getBufferedStreamEcg().broadcast
      ..bufferedFilterStreamEcg = _getBufferedFilterStreamEcg().broadcast
      ..bufferedPlotStreamEcg = _getBufferedPlotStreamEcg().broadcast
      ..bufferedPlotRTStreamEcg = _getBufferedPlotRTStreamEcg().broadcast;

    _isMeasuring = true;
    _isPaused = false;
  }

  /// [pauseMeasurementEcg] pause measurements for Ecg.
  Future<void> pauseMeasurementEcg() async {
    _isPaused = !_isPaused;
  }

  /// [stopMeasurementEcg] sends stop_ecg_packet to device.
  Future<void> stopMeasurementEcg() async {
    _isMeasuring = false;
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetStopEcg}');
    await control.write(Commands.packetStopEcg);
  }

  Future<bool> _checkValidationEcg() async {
    final confirm = points[Characteristics.DEVICE_CONTROL_CONFIRM];
    if (confirm == null) return false;

    final confirmValue = await confirm.read();
    if (confirmValue.isEmpty) return false;
    if (confirmValue[0] == 2) return false;
    return true;
  }

  Future<void> _sendCommandReadEcg() async {
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetReadEcg}');
    await control.write(Commands.packetReadEcg);
  }

  Stream<List<int>> _getStreamEcg() async* {
    final oneChannel = points[Characteristics.ONE_CHANNEL];
    if (oneChannel == null) return;
    await oneChannel.setNotifyValue(true);
    await for (final res in oneChannel.value) {
      if (!_isMeasuring || device == null) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getDecompressedStreamEcg() async* {
    final events = streams.rawStreamEcg;
    if (events == null) return;
    await for (final event in events) {
      yield lzma.decode(event);
    }
  }

  Stream<List<int>>? get _events {
    if (isCompressedEcg) return streams.decompressedStreamEcg;
    return streams.rawStreamEcg;
  }

  Stream<List<double>> _getRecentStreamEcg() async* {
    final events = _events;
    if (events == null) return;
    final res = <double>[];
    await for (final event in events) {
      if (event.isEmpty) continue;
      final itemCount = event.length ~/ 3;
      res.clear();
      for (var i = 0; i < itemCount; i++) {
        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );
        res.add(raw);
      }
      yield res;
    }
  }

  Stream<LeadState> _getLeadStateStream() async* {
    final leadState = points[Characteristics.LEAD_STATE];
    if (leadState == null) return;
    await leadState.setNotifyValue(true);
    await for (final res in leadState.value) {
      if (res.length < 2) continue;
      yield LeadState(
        ra: res[3] & 0x01 == 0x01,
        rl: true,
        la: res[3] >> 1 & 0x01 == 0x01,
        ll: true,
        v1: true,
        v2: true,
        v3: true,
        v4: true,
        v5: true,
        v6: true,
      );
    }
  }

  Stream<int> _getHeartRateStream() async* {
    final events = streams.recentStreamEcg;
    if (events == null) return;

    await for (final event in events) {
      if (event.isEmpty) continue;

      for (var i = 0; i < event.length; i++) {
        final output = _adaptiveHR.calculateHR(
          event[i],
          true,
        );

        if (output > 0) yield output;
      }
    }
  }

  Stream<Channels> _getBufferedStreamEcg() async* {
    final events = streams.recentStreamEcg;
    if (events == null) return;
    final res = List.generate(12, (index) => <double>[]);
    await for (final event in events) {
      res[1].addAll(event);
      if (res[1].length > maxBufferStreamEcg) {
        res[1].removeRange(0, res[1].length - lengthBufferStreamEcg);
      }
      yield Channels.fromList(res);
    }
  }

  Stream<Channels> _getBufferedFilterStreamEcg() async* {
    final events = streams.bufferedStreamEcg;
    if (events == null) return;

    await for (final event in events) {
      // ignore: lines_longer_than_80_chars
      for (var i = event.l2.length - _sampleCountEcg; i < event.l2.length; i++) {
        event.l2[i] = _dataFilter.applyFilter(1, event.l2[i], true);
      }
      yield event;
    }
  }

  Stream<PlotChannels> _getBufferedPlotStreamEcg() async* {
    // ignore: lines_longer_than_80_chars
    final events = isFilteredEcg ? streams.bufferedStreamEcg : streams.bufferedFilterStreamEcg;

    if (events == null) return;

    final resList = List.generate(12, (index) => <FlSpot>[]);
    final res = PlotChannels.fromList(resList, 0);
    var x = 0.0;

    await for (final event in events) {
      // ignore: lines_longer_than_80_chars
      for (var i = event.l2.length - _sampleCountEcg; i < event.l2.length; i++) {
        final j = i - event.l2.length + _sampleCountEcg;
        res.l2.add(FlSpot(x + j, event.l2[i]));
      }

      if (res.l2.length > event.l2.length) {
        res.l2.removeRange(0, res.l2.length - event.l2.length);
      }

      yield res;
      x += _sampleCountEcg;
    }
  }

  Stream<PlotChannels> _getBufferedPlotRTStreamEcg() async* {
    final events = streams.bufferedPlotStreamEcg;

    if (events == null) return;

    double skip(double x) {
      final newX = x % screenBufferStreamingEcg;
      final newXSkip = newX - newX % skipSampleEcg;
      return newXSkip;
    }

    FlSpot f(FlSpot original) {
      return FlSpot(
        skip(original.x),
        original.y,
      );
    }

    final l2 = List<FlSpot>.generate(
      screenBufferStreamingEcg ~/ skipSampleEcg,
      (index) => FlSpot.nullSpot,
      growable: false,
    );

    final resList = List.generate(
      12,
      (index) => (index == 1) ? l2 : <FlSpot>[],
    );

    final res = PlotChannels.fromList(resList, 0);
    var x = 0;

    await for (final event in events) {
      // ignore: lines_longer_than_80_chars
      for (var i = event.l2.length - _sampleCountEcg; i < event.l2.length; i++) {
        final j = i - event.l2.length + _sampleCountEcg;

        if (skipSampleEcg != 1) {
          if ((x + j) % skipSampleEcg != 0) continue;
        }

        final index = skip(x + j.toDouble());

        res.l2[index ~/ skipSampleEcg] = f(event.l2[i]);

        if (refreshBufferEcg > 0) {
          final refreshedIndex = skip(x + j.toDouble() + refreshBufferEcg);

          res.l2[refreshedIndex ~/ skipSampleEcg] = FlSpot.nullSpot;
        }
      }

      x += _sampleCountEcg;

      if (skipSampleEcg == 1 || x ~/ _sampleCountEcg % skipSampleEcg == 0) {
        yield res.copyWith(
          pivot: (x % screenBufferStreamingEcg).toDouble(),
        );
      }
    }
  }
}
