// ignore_for_file: public_member_api_docs,
// ignore_for_file: use_late_for_private_fields_and_variables

import 'dart:developer';

import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_support_for_wearcardiot/healthrian_ble_support_for_wearcardiot.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

mixin Oxi {
  /// [maxBufferStreamOxi] decides a maximum length of buffered Stream.
  ///
  /// For 40 seconds buffer and 100 sampling rate,
  /// It uses 40 * 1000.
  int maxBufferStreamOxi = 40 * 100;

  /// [lengthBufferStreamOxi] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  ///
  /// For 30 seconds buffer and 100 sampling rate,
  /// It uses 30 * 100.
  int lengthBufferStreamOxi = 30 * 100;

  /// [screenBufferStreamingOxi] decides a length of buffered Stream,
  /// which is shown to screen.
  ///
  /// For 5 seconds buffer and 100 sampling rate,
  /// It uses 5 * 100 by default.
  int screenBufferStreamingOxi = 5 * 100;

  /// [refreshBufferOxi] decides a length to be hidden after refreshed.
  ///
  /// For 0.5 seconds range to be hidden and 250 sampling rate,
  /// It use 100 ~/ 2 by default.
  int refreshBufferOxi = 100 ~/ 2;

  /// [isFilteredOxi] decides whether pulse oxi signal is filtered.
  bool isFilteredOxi = true;

  /// [isCompressedOxi] decides whether pulse oxi signal is compressed.
  ///
  /// If it is `true`, pulse oxi signal is decompressed by LZMA algorithm.
  bool isCompressedOxi = false;

  int _channelCountOxi = 2;
  int _sampleCountOxi = 30;

  int get _totalCountOxi => _channelCountOxi * _sampleCountOxi;

  bool _isPaused = false;

  bool _isMeasuring = false;

  /// [skipSampleOxi] is the number of samples to be skipped.
  int skipSampleOxi = 1;

  /// [skipPacketOxi] is the number of packet to be skipped.
  int skipPacketOxi = 1;
}

extension BLEOxi on BLEManager {
  /// [startMeasurementOxi] sends read_oxi_packet to device,
  ///  and then, it registers all oxi streams of [streams].
  Future<void> startMeasurementOxi() async {
    await _getConfigurationOxi();
    await _sendCommandReadOxi();

    // final isRunnable = await _checkValidationOxi();
    // if (!isRunnable) return;

    streams
      ..rawStreamOxi = _getStreamOxi().broadcast
      ..decompressedStreamOxi = _getDecompressedStreamOxi().broadcast
      ..recentStreamOxi = _getRecentStreamOxi().broadcast
      ..bufferedStreamOxi = _getBufferedStreamOxi().broadcast
      ..bufferedFilterStreamOxi = _getBufferedFilterStreamOxi().broadcast
      ..bufferedPlotStreamOxi = _getBufferedPlotStreamOxi().broadcast
      ..bufferedPlotRTStreamOxi = _getBufferedPlotRTStreamOxi().broadcast;

    _isPaused = false;
    _isMeasuring = true;
  }

  /// [pauseMeasurementOxi] pause measurements for Pulse Oxi.
  Future<void> pauseMeasurementOxi() async {
    _isPaused = !_isPaused;
  }

  /// [stopMeasurementOxi] sends stop_oxi_packet to device.
  Future<void> stopMeasurementOxi() async {
    _isMeasuring = false;
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetStopOxi}');
    await control.write(Commands.packetStopOxi);
  }

  Future<bool> _checkValidationOxi() async {
    final confirm = points[Characteristics.DEVICE_CONTROL_CONFIRM];
    if (confirm == null) return false;

    final confirmValue = await confirm.read();
    if (confirmValue.isEmpty) return false;
    if (confirmValue[0] == 2) return false;
    return true;
  }

  Future<void> _sendCommandReadOxi() async {
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetReadOxi}');
    await control.write(Commands.packetReadOxi);
  }

  Future<void> _getConfigurationOxi() async {
    final configuration = points[Characteristics.PULSE_OXIMETER_CONFIGURATION];
    if (configuration == null) return;

    final configurationValue = await configuration.read();

    if (configurationValue.length < 2) return;

    _channelCountOxi = configurationValue[0];
    _sampleCountOxi = configurationValue[1];
  }

  Stream<List<int>> _getStreamOxi() async* {
    final continuous = points[Characteristics.PULSE_OXIMETER_CONTINUOUS];
    if (continuous == null) return;
    await continuous.setNotifyValue(true);
    await for (final res in continuous.value) {
      if (!_isMeasuring || device == null) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getDecompressedStreamOxi() async* {
    final events = streams.rawStreamOxi;
    if (events == null) return;
    await for (final event in events) {
      yield isCompressedOxi ? lzma.decode(event) : event;
    }
  }

  Stream<List<double>> _getRecentStreamOxi() async* {
    final events = streams.rawStreamOxi;
    if (events == null) return;
    final res = <double>[];
    await for (final event in events) {
      if (event.isEmpty) continue;
      res.clear();
      for (var i = 0; i < _totalCountOxi; i++) {
        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );
        res.add(raw);
      }
      yield res;
    }
  }

  Stream<Channels> _getBufferedStreamOxi() async* {
    final events = streams.recentStreamOxi;
    if (events == null) return;
    final res = List.generate(12, (index) => <double>[]);
    await for (final event in events) {
      for (var i = 0; i < event.length; i++) {
        res[i % _channelCountOxi].add(event[i]);
      }
      if (res[0].length > maxBufferStreamOxi) {
        for (var j = 0; j < _channelCountOxi; j++) {
          res[j].removeRange(0, res[j].length - maxBufferStreamOxi);
        }
      }
      yield Channels.fromList(res);
    }
  }

  Stream<Channels> _getBufferedFilterStreamOxi() async* {
    final events = streams.bufferedStreamOxi;
    if (events == null) return;

    await for (final event in events) {
      // ignore: lines_longer_than_80_chars
      for (var i = event.l2.length - _sampleCountOxi; i < event.l2.length; i++) {
        for (var j = 0; j < _channelCountOxi; j++) {
          // ignore: lines_longer_than_80_chars
          // event.toList()[j] = _dataFilter.applyFilter(j, event.toList()[j], true);
        }
      }
      yield event;
    }
  }

  Stream<PlotChannels> _getBufferedPlotStreamOxi() async* {
    // ignore: lines_longer_than_80_chars
    final events = isFilteredOxi ? streams.bufferedStreamOxi : streams.bufferedFilterStreamOxi;

    if (events == null) return;

    final resList = List.generate(12, (index) => <FlSpot>[]);
    final res = PlotChannels.fromList(resList, 0);
    final r = res.toList();
    var x = 0.0;
    await for (final event in events) {
      final e = event.toList();
      for (var i = e[0].length - _sampleCountOxi; i < e[0].length; i++) {
        final j = i - e[0].length + _sampleCountOxi;

        for (var k = 0; k < _sampleCountOxi; k++) {
          r[k].add(FlSpot(x + j, e[k][i]));
        }
      }

      if (r[0].length > e[0].length) {
        for (var k = 0; k < _sampleCountOxi; k++) {
          r[k].removeRange(0, r[k].length - e[k].length);
        }
      }

      yield res;
      x += _sampleCountOxi;
    }
  }

  Stream<PlotChannels> _getBufferedPlotRTStreamOxi() async* {
    final events = streams.bufferedPlotStreamOxi;

    if (events == null) return;

    double skip(double x) {
      final newX = x % screenBufferStreamingOxi;
      final newXSkip = newX - newX % skipSampleOxi;
      return newXSkip;
    }

    FlSpot f(FlSpot original) {
      return FlSpot(
        skip(original.x),
        original.y,
      );
    }

    // final l1 = List<FlSpot>.generate(
    //   screenBufferStreamingOxi ~/ skipSampleOxi,
    //   (index) => FlSpot.nullSpot,
    //   growable: false,
    // );
    //
    // final l2 = List<FlSpot>.generate(
    //   screenBufferStreamingOxi ~/ skipSampleOxi,
    //   (index) => FlSpot.nullSpot,
    //   growable: false,
    // );

    final resList = List.generate(
      12,
      (index) {
        if (index < _channelCountOxi) {
          return List<FlSpot>.generate(
            screenBufferStreamingOxi ~/ skipSampleOxi,
            (index) => FlSpot.nullSpot,
            growable: false,
          );
        }
        return <FlSpot>[];
      },
    );

    final res = PlotChannels.fromList(resList, 0);
    final r = res.toList();
    var x = 0;

    await for (final event in events) {
      final e = event.toList();
      // ignore: lines_longer_than_80_chars
      for (var i = event.l2.length - _sampleCountOxi; i < event.l2.length; i++) {
        final j = i - event.l2.length + _sampleCountOxi;

        if (skipSampleOxi != 1) {
          if ((x + j) % skipSampleOxi != 0) continue;
        }

        final index = skip(x + j.toDouble());

        for (var k = 0; k < _sampleCountOxi; k++) {
          r[k][index ~/ skipSampleOxi] = f(e[k][i]);
        }

        if (refreshBufferOxi > 0) {
          final refreshedIndex = skip(x + j.toDouble() + refreshBufferOxi);

          for (var k = 0; k < _sampleCountOxi; k++) {
            r[k][refreshedIndex ~/ skipSampleOxi] = FlSpot.nullSpot;
          }
        }
      }

      x += _sampleCountOxi;

      if (skipSampleOxi == 1 || x ~/ _sampleCountOxi % skipSampleOxi == 0) {
        yield res.copyWith(
          pivot: (x % screenBufferStreamingOxi).toDouble(),
        );
      }
    }
  }
}
